/*
 * Simple PPP login front-end.
 * Should be made setuid, and setuid bit should be removed from pppd.
 *
 * Copyright (C) 1999, Caldera (Deutschalnd) GmbH
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/syslog.h>
#include <dirent.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <getopt.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>

/*
 * Location of ppplogin configuration file
 */
#define PPPLOGIN_CONF	"/etc/ppp/ppplogin.conf"

/*
 * Location of log file
 */
#define PPPLOGIN_LOG	"/var/log/dialin"

/*
 * Define this if your passwd lives in a special location.
 * Otherwise we try /bin/passwd and /usr/bin/passwd
 */
#if 0
#define PATH_BINPASSWD	"/some/place/passwd"
#endif

struct pppconf {
	struct in_addr	locip;
	struct in_addr	remip;
	char *		command;
	int		chpass;
};

static void		execute(struct pppconf *, const char *, const char *);
static void		password(struct pppconf *, const char *, const char *);
static struct pppconf *	getconfig(const char *);
static const char *	getline();
static void		log(const char *, ...);
static void		fatal(const char *, ...);
static void		usage(int);

int
main(int argc, char **argv)
{
	struct pppconf	*cfg;
	char		hostname[256];
	const char	*line, *user;
	int		c;

	if (!isatty(0) || !isatty(1) || !isatty(2)) {
		printf("Not a tty.\n");
		exit(1);
	}

	while ((c = getopt(argc, argv, "h")) != -1) {
		switch (c) {
		case '?':
			usage(0);
		default:
			usage(1);
		}
	}

	/* Open syslog */
	openlog("ppplogin", 0, LOG_DAEMON);

	/* Obtain the user name */
	if (!(user = getlogin()))
		fatal("You don't exist, go away");

	/* Obtain the tty name */
	line = getline();

	if (gethostname(hostname, sizeof(hostname)) < 0)
		fatal("Unable to obtain hostname: %m\n");

	/* Obtain the user's ppp config */
	cfg = getconfig(user);

	/* Make sure files get closed on exec */
	for (c = 3; c < getdtablesize(); c++)
		fcntl(c, F_SETFD, 1);

	printf("Welcome to %s!\n"
	       "You are calling from %s\n", hostname, line);

	if (!strncmp(line, "/dev/ttyS", 9)
	 || !strncmp(line, "/dev/ttyI", 9)) {
		/* Okay, coming in via dialup link */
		execute(cfg, user, line);
	} else {
		/* Login via console or telnet: change password */
		password(cfg, user, line);
	}

	return 0;
}

static void
execute(struct pppconf *cfg, const char *user, const char *line)
{
	time_t	now;
	char	*argv[32], *sp, addrspec[64];
	int	argc;

	if (!cfg->command)
		fatal("Missing command for user %s\n", user);
	if (!cfg->locip.s_addr || !cfg->remip.s_addr)
		fatal("Missing local or remote IP address for user %s\n", user);

	printf("My IP address is '%s'.\n", inet_ntoa(cfg->locip));
	printf("Your IP address is '%s'.\n", inet_ntoa(cfg->remip));
	printf("LaunchingPPP\n");

	argc = 0;
	sp = strtok(cfg->command, " \t");
	do {
		argv[argc++] = sp;
	} while ((sp = strtok(NULL, " \t")) != 0);

	/* Add locip:remip */
	strcpy(addrspec, inet_ntoa(cfg->locip));
	strcat(addrspec, ":");
	strcat(addrspec, inet_ntoa(cfg->remip));
	argv[argc++] = addrspec;

	/* Terminate argument vector for execv */
	argv[argc] = NULL;

	if (setuid(0) < 0)
		fatal("Unable to become super user: %m\n");

	time(&now);
	log("Dialin on %s on %s from %s\n", line, ctime(&now), user);

	execv(argv[0], argv);
	fatal("Unable to execute %s: %m", argv[0]);
}

static void
password(struct pppconf *cfg, const char *user, const char *line)
{
	char	*argv[2];

	if (!cfg->chpass)
		fatal("Sorry, you are not allowed to change "
			"your password.\n");
	/* Drop root privs */
	if (setuid(getuid()) < 0)
		fatal("Unable to drop root privilege.\n");

	printf("You can now change your password.\n");
	argv[0] = "passwd";
	argv[1] = NULL;

#ifdef PATH_BINPASSWD
	execv(PATH_BINPASSWD, argv);
#else
	execv("/bin/passwd", argv);
	execv("/usr/bin/passwd", argv);
#endif
	fatal("Unable to execute passwd program: %m\n");
}

static struct pppconf *
getconfig(const char *user)
{
	static struct pppconf	cfg;
	FILE			*fp;
	char			buffer[512], *kwd, *arg, *sp, *errmsg;
	int			ignore = 1, line = 0, found = 0;

	if (!(fp = fopen(PPPLOGIN_CONF, "r")))
		fatal("Unable to open %s: %m\n", PPPLOGIN_CONF);

	memset(&cfg, 0, sizeof(cfg));
	while (fgets(buffer, sizeof(buffer), fp) != NULL) {
		++line;
		if (!(kwd = strtok(buffer, " \t\n")))
			continue;
		if (kwd[0] == '#')
			continue;
		if (!strcmp(kwd, "defaults")) {
			ignore = 0;
			continue;
		}
		arg = strtok(NULL, "\n#");
		while (isspace(*arg))
			arg++;
		sp = arg + strlen(arg);
		while (sp > arg && isspace(sp[-1]))
			*--sp = '\0';
		if (!strcmp(kwd, "user")) {
			errmsg = "Missing user name";
			if (!arg)
				goto syntax;
			if (found)
				fatal("Duplicate entry for user %s in %s\n",
					user, PPPLOGIN_CONF);
			ignore = 1;
			if (!strcasecmp(user, arg)) {
				found = 1;
				ignore = 0;
			}
			continue;
		} else if (ignore)
			continue;
		if (!strcmp(kwd, "local-ip") || !strcmp(kwd, "remote-ip")) {
			struct in_addr	addr;

			errmsg = "Missing or bad IP address";
			if (!arg || !inet_aton(arg, &addr))
				goto syntax;
			if (kwd[0] == 'l')
				cfg.locip = addr;
			else
				cfg.remip = addr;
		} else
		if (!strcmp(kwd, "change-password")) {
			errmsg = "Missing or bad allow/deny";
			if (!arg)
				goto syntax;
			if (!strcasecmp(arg, "allow"))
				cfg.chpass = 1;
			else if (!strcasecmp(arg, "deny"))
				cfg.chpass = 0;
			else
				goto syntax;
		} else
		if (!strcmp(kwd, "command")) {
			errmsg = "Missing command";
			if (!arg)
				goto syntax;
			cfg.command = strdup(arg);
		} else {
			errmsg = "Unknown keyword";
			goto syntax;
		}
	}
	if (!found)
		fatal("You are not allowed to use ppplogin\n");

	fclose(fp);
	return &cfg;

syntax:
	fatal("%s in %s, line %d\n", errmsg, PPPLOGIN_CONF, line);
	return 0;
}

static const char *
getline(void)
{
	static char	devname[256];
	struct stat	stb;
	struct DIR	*dir;
	struct dirent	*de;

	if (fstat(0, &stb) < 0)
		fatal("Unable to stat stdin: %m\n");

	if (!(dir = opendir("/dev")))
		fatal("Unable to open /dev: %m\n");

	while ((de = readdir(dir)) != 0) {
		if (de->d_ino == stb.st_ino) {
			snprintf(devname, sizeof(devname),
					"/dev/%s", de->d_name);
			return devname;
		}
	}

	fatal("Unable to determine tty.\n");
	/* notreached */
	return 0;
}

static void
fatal(const char *msg, ...)
{
	va_list		ap;

	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	vsyslog(LOG_WARNING, msg, ap);
	va_end(ap);
	exit(1);
}

static void
log(const char *msg, ...)
{
	static FILE	*logfp = 0;
	va_list		ap;

	if (!logfp && !(logfp = fopen(PPPLOGIN_LOG, "a")))
		fatal("Unable to open %s: %m\n", PPPLOGIN_LOG);
	va_start(ap, msg);
	vfprintf(logfp, msg, ap);
	va_end(ap);
}

static void
usage(int exval)
{
	fprintf(stderr, "usage: ppplogin [-h]\n");
	exit(exval);
}
